/*
 * Copyright 1997 Vectaport Inc.
 * Copyright 1994, 1995 Vectaport Inc., Cartoactive Systems
 *
 * Permission to use, copy, modify, distribute, and sell this software and
 * its documentation for any purpose is hereby granted without fee, provided
 * that the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the names of the copyright holders not be used in
 * advertising or publicity pertaining to distribution of the software
 * without specific, written prior permission.  The copyright holders make
 * no representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING
 * FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
 * NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 * 
 */

#include <Map/mapcov.h>
#include <Map/mapdb.h>
#include <Map/mapfclass.h>
#include <Map/mapfeature.h>
#include <Map/maplib.h>
#include <Map/mapprim.h>

#include <MapUnidraw/mapcmds.h>
#include <MapUnidraw/mapclasses.h>
#include <MapUnidraw/mapcomps.h>
#include <MapUnidraw/mapeditor.h>
#include <MapUnidraw/mapviewer.h>
#include <MapUnidraw/mapviews.h>
#include <MapUnidraw/mapviewstate.h>

#include <OverlayUnidraw/ovcatalog.h>

#include <UniIdraw/idarrows.h>

#include <Unidraw/globals.h>
#include <Unidraw/Commands/command.h>

#include <Unidraw/Components/text.h>

#include <Unidraw/Graphic/damage.h>
#include <Unidraw/Graphic/ellipses.h>
#include <Unidraw/Graphic/graphic.h>
#include <Unidraw/Graphic/lines.h>
#include <Unidraw/Graphic/polygons.h>

#include <Unidraw/catalog.h>
#include <Unidraw/viewer.h>
#include <Unidraw/unidraw.h>

#include <InterViews/transformer.h>

#include <Attribute/attribute.h>
#include <Attribute/attrlist.h>
#include <Attribute/attrvalue.h>

#include <iostream.h>


/****************************************************************************/

FullGraphic* MapFeatureView::_afv_gs = nil;
FullGraphic* MapFeatureView::_til_gs = nil;
FullGraphic* MapFeatureView::_lfv_gs = nil;
FullGraphic* MapFeatureView::_pfv_gs = nil;
FullGraphic* MapFeatureView::_tfv_gs = nil;

#define EPS10 1.0e-10

void project(PJ* proj, int npts, const float* fx, const float* fy, int* ix, int* iy, float scale = 1.0, Transformer* t=nil) {
    UV data;
    for (int i=0; i<npts; i++) {
	if (proj) {
	    data.u = fx[i] * DEG_TO_RAD;
	    data.v = fy[i] * DEG_TO_RAD;
#if 0
	    if (t && !t->identity()) {
	      float fxf = data.u;
	      float fyf = data.v;
	      t->Transform(fxf, fyf, fxf, fyf);
	      data.u = fxf;
	      data.v = fyf;
	    }
#endif
	    data = pj_fwd(data, proj);
	    if (data.u == HUGE_VAL) {
		/* cerr << "projection error\n"; */
		data.u = fx[i] * DEG_TO_RAD;
		data.v = fy[i] * DEG_TO_RAD;
		if (fx[i]==90.0)
		    data.u -= 2 * EPS10;
		if (fx[i]==-90.0)
		    data.u += 2 * EPS10;
		if (fy[i]==90.0)
		    data.v -= 2 * EPS10;
		if (fy[i]==-90.0)
		    data.v += 2 * EPS10;
		data = pj_fwd(data, proj);
		if (data.u == HUGE_VAL)
		    cerr << "projection error again\n";
	    }

	    if (t && !t->identity()) {
	      float fxf = data.u;
	      float fyf = data.v;
	      t->Transform(fxf, fyf, fxf, fyf);
	      data.u = fxf;
	      data.v = fyf;
	    }

	    ix[i] = round(data.u);
	    iy[i] = round(data.v);
	} else {
	    float fxf = fx[i];
	    float fyf = fy[i];
	    if (t && !t->identity()) 
	      t->Transform(fxf, fyf, fxf, fyf);
	    ix[i] = round(fxf * scale);
	    iy[i] = round(fyf * scale);
	}
    }
}

/****************************************************************************/

ClassId MapFeatureView::GetClassId () { return MAPFEATURE_VIEW; }

boolean MapFeatureView::IsA (ClassId id) {
    return MAPFEATURE_VIEW == id || OverlayView::IsA(id);
}

MapFeatureView::MapFeatureView(MapFeatureComp* comp) : OverlayView()
{
    SetSubject(comp);
    _orig_gs = nil;
}

MapFeatureView::~MapFeatureView() {
  if (GetGraphic() == GetGraphicComp()->GetGraphic())
    GetGraphicComp()->SetGraphic(nil);
}

void MapFeatureView::Update() {
    MapFeatureComp* comp = (MapFeatureComp*)GetGraphicComp();
    MapFeature* feature = comp->feature();
    ComponentView* topview = this;
    while (!topview->IsA(MAP_IDRAW_VIEW) && topview->GetParent())
      topview = topview->GetParent();
    MapEditor* mapeditor = topview->IsA(MAP_IDRAW_VIEW) ? 
      ((MapIdrawView*)topview)->GetMapEditor() : nil;
    MapViewState* mapviewstate = mapeditor ? mapeditor->GetMapViewState() : nil;

    PJ* proj = mapviewstate ? mapviewstate->defproj() : nil;
    Graphic* gr = GetGraphic();  
    Transformer* grtr = nil;
    if (mapviewstate) {
      gr->concatGS(mapviewstate->themegs(feature), gr, gr);
      if (mapviewstate->normtrans())
	gr->SetTransformer(new Transformer(mapviewstate->deftrans()));
      else {
	if (gr->GetTransformer()) gr->SetTransformer(nil);
	grtr = mapviewstate->deftrans();
      }
    } else
      gr->concatGS(stdgraphic, gr, gr);

    AttributeList* al = comp->GetAttributeList();
    if (al->GetAttr("sensitive")) {
      AttributeValue* av = al->GetAttr("sensitive")->Value();
      if (av && av->is_false())
	Desensitize();
      else
	Sensitize();
    }
    
    switch(feature->type()) {
    case MapFeature::PointType:
	{
	    float mag= 1.0;
	    if (GetViewer())
	      mag = GetViewer()->GetMagnification();
	    MapPoint* point = feature->point_primitive();
	    if (!point) return;
	    int ix[1];
	    int iy[1];
	    const float fx[1] = { point->xorigin() };
	    const float fy[1] = { point->yorigin() };
	    project(proj, 1, fx, fy, ix, iy, feature->unprojscale(), grtr);
	    FixSize();
	    IncurDamage(gr);
	    float a00, a01, a10, a11, a20, a21;
	    if (mapviewstate && mapviewstate->deftrans()) 
	      mapviewstate->deftrans()->matrix(a00,a01,a10,a11,a20,a21);
	    else {
	      a00 = a11 = 1.0;
	      a01 = a10 = a20 = a21 = 0.0;
	    }
	    int radius;
	    if (feature->feature_class() && feature->feature_class()->tiled())
	      radius = (int)(4 * (1.0 / (a00 * 2.0)));
	    else
	      radius = (int)(4 * (1.0 / a00 * 0.5));
	    ((SF_Ellipse*)gr)->SetOriginal(ix[0], iy[0],
					   round(radius/mag),
					   round(radius/mag));
	    IncurDamage(gr);
	}
	break;
    case MapFeature::LineType:
	{
	    MapEdge* edge = feature->edge_primitive();
	    if (!edge) return;
	    int npts = edge->npts();
	    int* ix = new Coord[npts];
	    int* iy = new Coord[npts];
	    const float* fx = edge->xpoints();
	    const float* fy = edge->ypoints();
	    project(proj, npts, fx, fy, ix, iy, feature->unprojscale(), grtr);
	    IncurDamage(gr);
	    MultiLineObj* mlo = new MultiLineObj(ix, iy, npts);
	    ((SF_MultiLine*)gr)->SetOriginal(mlo);
	    IncurDamage(gr);
	}
	break;
    case MapFeature::AreaType:
	{
	    MapFace* face = feature->face_primitive();
	    if (!face) return;
	    int npts = face->npts();
	    int* ix = new Coord[npts];
	    int* iy = new Coord[npts];
	    const float* fx = face->xpoints();
	    const float* fy = face->ypoints();
	    project(proj, npts, fx, fy, ix, iy, feature->unprojscale(), grtr);
	    IncurDamage(gr);
	    MultiLineObj* mlo = new MultiLineObj(ix, iy, npts);
	    ((SF_Polygon*)gr)->SetOriginal(mlo);
	    IncurDamage(gr);
	}
	break;
    case MapFeature::TextType:
	{
	    float mag= GetViewer()->GetMagnification();
	    MapText* text = feature->text_primitive();
	    if (!text) return;
	    FixSize(.6);
	    IncurDamage(gr);
	    if (comp->pointcomp()) { // companion point comp for text comp
	      MapFeatureComp* pfc = comp->pointcomp();
	      Iterator cit;
	      pfc->First(cit);
	      MapFeatureView* pfv = (MapFeatureView*)pfc->FindView(GetViewer());
	      Graphic* el = pfv->GetGraphic();

	      // align text graphic to point graphic
	      el->Align(10, gr, 9);  // abut right
	      el->Align(14, gr, 14); // vert centers

	      // make companion point graphic invisible
	      el->SetPattern(psnonepat);
	      el->SetBrush(psnonebr);

	      // scaling kludge for DCW text:
	      float cx, cy;
	      gr->GetCenter(cx, cy);
	      if (feature->feature_class()->coverage()->library()->tiled())
		gr->Scale(200,200,cx,cy);
	      else
		gr->Scale(14000,14000,cx,cy);
	    }

	    IncurDamage(gr);
	}
	break;
    case MapFeature::JunctionType:
	{
	    float mag= 1.0;
	    if (GetViewer())
	      mag = GetViewer()->GetMagnification();
	    MapNode* node = feature->node_primitive();
	    if (!node) return;
	    int ix[1];
	    int iy[1];
	    const float fx[1] = { node->xorigin() };
	    const float fy[1] = { node->yorigin() };
	    project(proj, 1, fx, fy, ix, iy, feature->unprojscale(), grtr);
	    FixSize();
	    IncurDamage(gr);
	    float a00, a01, a10, a11, a20, a21;
	    if (mapviewstate && mapviewstate->deftrans()) 
	      mapviewstate->deftrans()->matrix(a00,a01,a10,a11,a20,a21);
	    else {
	      a00 = a11 = 1.0;
	      a01 = a10 = a20 = a21 = 0.0;
	    }
	    int radius;
	    if (feature->feature_class() && feature->feature_class()->tiled())
	      radius = (int)(4 * (1.0 / (a00 * 2.0)));
	    else
	      radius = (int)(4 * (1.0 / a00 * 0.5));
	    ((SF_Ellipse*)gr)->SetOriginal(ix[0], iy[0],
					   round(radius/mag),
					   round(radius/mag));
	    IncurDamage(gr);
	}
	break;
    case MapFeature::RasterType:
    case MapFeature::DemType:
    default:
	cerr << "Unsupported (as of yet) map feature\n";
	break;
    }
}


Graphic* MapFeatureView::GetGraphic() {
    Graphic* graphic = GraphicView::GetGraphic();
    if (graphic != nil) return graphic;

    MapFeatureComp* comp = (MapFeatureComp*)GetGraphicComp();
    MapFeature* feature = comp->feature();
    Graphic* compgr = comp->GetGraphic();
    Graphic* gs = compgr ? compgr : nil;
    Graphic* gr = nil;

    switch(feature->type()) {
    case MapFeature::PointType:
	{
	    MapPoint* point = feature->point_primitive();
	    if (!point) return nil;
	    int ix[1];
	    int iy[1];
	    gr = new SF_Ellipse (ix[0], iy[0], 1, 1, gs);
	}
	break;
    case MapFeature::LineType:
	{
	    MapEdge* edge = feature->edge_primitive();
	    if (!edge) return nil;
	    int npts = edge->npts();
	    int ix[npts];
	    int iy[npts];
//	    gs->SetPattern(psnonepat);
	    gr = new ArrowMultiLine (ix, iy, npts, false, false, 1.0, gs);
	}
	break;
    case MapFeature::AreaType:
	{
	    MapFace* face = feature->face_primitive();
	    if (!face) return nil;
	    int npts = face->npts();
	    int ix[npts];
	    int iy[npts];
	    gr = new SF_Polygon(ix, iy, npts, gs);
	}
	break;
    case MapFeature::TextType:
	{
	    MapText* text = feature->text_primitive();
	    if (!text) return nil;
	    boolean alloc_gs = false;
  	    if (gs) 
	      gs->SetFont(stdgraphic->GetFont());
	    else {
	      alloc_gs = true;
	      gs = new FullGraphic(stdgraphic);
	      gs->FillBg(0);
	      gs->SetPattern(psnonepat);
	    }
	    gr = new TextGraphic(text->text(), gs);
	    if (alloc_gs) delete gs;
	}
	break;
    case MapFeature::JunctionType:
	{
	    MapNode* node = feature->node_primitive();
	    if (!node) return nil;
	    int ix[1];
	    int iy[1];
	    gr = new SF_Ellipse (ix[0], iy[0], 1, 1, gs);
	}
	break;
    case MapFeature::RasterType:
    case MapFeature::DemType:
    default:
	cerr << "Unsupported (as of yet) map feature\n";
	break;
    }

    SetGraphic(gr);
    comp->graphic(gr);
    gr->SetTag(this);
    return gr;
}

Graphic* MapFeatureView::HighlightGraphic() {
    if (!_afv_gs) {
	_afv_gs = new FullGraphic();
	_afv_gs->SetPattern(unidraw->GetCatalog()->ReadPattern("pattern", 4));
	_afv_gs->SetBrush(unidraw->GetCatalog()->FindBrush(0xffff, 1));
	_til_gs = new FullGraphic();
	_til_gs->SetPattern(unidraw->GetCatalog()->ReadPattern("pattern", 4));
	_til_gs->FillBg(false);
	_til_gs->SetBrush(unidraw->GetCatalog()->FindBrush(0xffff, 1));
	_lfv_gs = new FullGraphic();
	_lfv_gs->SetBrush(unidraw->GetCatalog()->FindBrush(0xfff0, 1));
	_pfv_gs = new FullGraphic();
	_pfv_gs->SetColors(unidraw->GetCatalog()->FindColor("red"), 
			   unidraw->GetCatalog()->FindColor("red"));
	_tfv_gs = new FullGraphic();
	_tfv_gs->SetColors(unidraw->GetCatalog()->FindColor("red"), 
			   unidraw->GetCatalog()->FindColor("red"));
	_tfv_gs->FillBg(false);
    }

    MapFeatureComp* comp = (MapFeatureComp*)GetGraphicComp();
    MapFeature* feature = comp->feature();
    switch (feature->type()) {
    case MapFeature::AreaType:
	if (feature->feature_class() && 
	    feature->feature_class()->name() &&
	    strcmp(feature->feature_class()->name(), "tileref")==0)
	    return _til_gs;
	else
	    return _afv_gs;
    case MapFeature::PointType:
	return _pfv_gs;
    case MapFeature::LineType:
	return _lfv_gs;
    case MapFeature::TextType:
	return _tfv_gs;
    case MapFeature::JunctionType:
	return _pfv_gs;
    default:
	return nil;
    }
    return nil;
}

void MapFeatureView::Highlight() {
  if (!_orig_gs)
    _orig_gs = GetGraphic()->Copy(); // save original gs
  OverlayView::Highlight();
}

void MapFeatureView::Unhighlight() {
    Graphic* gview = GetGraphic();
    Viewer* viewer = GetViewer();
    // just restore original gs, no need to paint with mapviewstate again
    //MapEditor* editor = (MapEditor*)viewer->GetEditor();
    //MapViewState* mapviewstate = editor->GetMapViewState();
    //MapFeatureComp* comp = (MapFeatureComp*)GetGraphicComp();
    //MapFeature* feature = comp->feature();
    Graphic* gs = _orig_gs; //mapviewstate->themegs(feature);
    
    viewer->GetDamage()->Incur(gview);
    Transformer* t = gview->GetTransformer();
    Ref(t);
    gview->SetTransformer(nil);
    if (gs)
      *gview = *gs;
    gview->SetTransformer(t);
    Unref(t);
    viewer->GetDamage()->Incur(gview);
}

/*****************************************************************************/

MapFClassView::MapFClassView(MapFClassComp* c)
: OverlaysView(c)
{
}

MapFClassView::~MapFClassView() {
}

ClassId MapFClassView::GetClassId() { return MAPFCLASS_VIEW; }

boolean MapFClassView::IsA(ClassId id) {
    return id == MAPFCLASS_VIEW || OverlaysView::IsA(id);
}

void MapFClassView::Interpret(Command* cmd) {
}

void MapFClassView::Uninterpret(Command* cmd) {
}

void MapFClassView::Update() { 
    MapFClassComp* comp = (MapFClassComp*)GetSubject();
    MapFeatureClass* fclass = comp->GetFeatureClass();
    if (!fclass || fclass->nfeats()==0) {
      if (fclass && fclass->path()) {
	comp->SetPathName(fclass->path());
	comp->SetFeatureClass(nil);
        delete fclass;
      }
	Component* newcomp;
	OverlayCatalog* catalog = (OverlayCatalog*)unidraw->GetCatalog();
	catalog->SetParent(comp);
	if (catalog->Retrieve(comp->GetPathName(), newcomp)) {
	  if (newcomp->IsA(MAPFCLASS_COMP)) {
	    MapFClassComp* fccomp = (MapFClassComp*)newcomp;
	    fclass = fccomp->GetFeatureClass();
	    comp->SetFeatureClass(fclass);
	    fccomp->SetFeatureClass(nil);
	    Iterator i;
	    for (fccomp->First(i); !fccomp->Done(i);) {
	      GraphicComp* fcomp = fccomp->GetComp(i);
	      fccomp->Remove(i);
	      comp->Append(fcomp);
	    }
	  }
	  delete newcomp;
	} else
	    cerr << "error in reading feature class from " << comp->GetPathName() << "\n";
	catalog->SetParent(nil);
    } 
    OverlaysView::Update();
}

/*****************************************************************************/

MapCovView::MapCovView(MapCovComp* c)
: OverlaysView(c)
{
}

MapCovView::~MapCovView() {
}

ClassId MapCovView::GetClassId() { return MAPCOV_VIEW; }

boolean MapCovView::IsA(ClassId id) {
    return id == MAPCOV_VIEW || OverlaysView::IsA(id);
}

MapFClassView* MapCovView::FindFClassView(MapFeatureClass* mapfc) {
    Iterator i;
    for (First(i); !Done(i); Next(i)) {
	MapFClassView* mfv = (MapFClassView*)GetView(i);
	if (((MapFClassComp*)mfv->GetGraphicComp())->GetFeatureClass() == mapfc)
	    return mfv;
    }
    return nil;
}

void MapCovView::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOW_CMD)) {
	MapFeatureClass* mapfc = (MapFeatureClass*)((MapCovComp*)GetGraphicComp())->GetCoverage()->feature_class(((MapShowCmd*)cmd)->GetFClassName());
	if (mapfc) {
	    MapFClassView* mfv = FindFClassView(mapfc);
	    if (mfv && mfv->Hidden()) {
		mfv->Show();
		mfv->Sensitize();
		GetViewer()->GetDamage()->Incur(GetGraphic());
		unidraw->Update();
	    }
	}
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	MapFeatureClass* mapfc = (MapFeatureClass*)((MapCovComp*)GetGraphicComp())->GetCoverage()->feature_class(((MapHideCmd*)cmd)->GetFClassName());
	if (mapfc) {
	    MapFClassView* mfv = FindFClassView(mapfc);
	    if (mfv && !mfv->Hidden()) {
		GetViewer()->GetDamage()->Incur(GetGraphic());
		mfv->Hide();
		mfv->Desensitize();
		unidraw->Update();
	    }
	}
    }
}

void MapCovView::Uninterpret(Command* cmd) {
}

/*****************************************************************************/

MapLibView::MapLibView(MapLibComp* c)
: OverlaysView(c)
{
}

MapLibView::~MapLibView() {
}

ClassId MapLibView::GetClassId() { return MAPLIB_VIEW; }

boolean MapLibView::IsA(ClassId id) {
    return id == MAPLIB_VIEW || OverlaysView::IsA(id);
}

MapCovView* MapLibView::FindCovView(MapCoverage* mapcov) {
    Iterator i;
    for (First(i); !Done(i); Next(i)) {
	MapCovView* mcv = (MapCovView*)GetView(i);
	if (((MapCovComp*)mcv->GetGraphicComp())->GetCoverage() == mapcov)
	    return mcv;
    }
    return nil;
}

void MapLibView::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOW_CMD)) {
	MapCoverage* mapcov = (MapCoverage*)((MapLibComp*)GetGraphicComp())->GetLibrary()->coverage(((MapShowCmd*)cmd)->GetCovName());
	if (mapcov) {
	    MapCovView* mcv = FindCovView(mapcov);
	    if (mcv)
		mcv->Interpret(cmd);
	}
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	MapCoverage* mapcov = (MapCoverage*)((MapLibComp*)GetGraphicComp())->GetLibrary()->coverage(((MapHideCmd*)cmd)->GetCovName());
	if (mapcov) {
	    MapCovView* mcv = FindCovView(mapcov);
	    if (mcv)
		mcv->Interpret(cmd);
	}
    }
    else OverlaysView::Interpret(cmd);
}

void MapLibView::Uninterpret(Command* cmd) {
}

/*****************************************************************************/

MapDbView::MapDbView(MapDbComp* c)
: OverlaysView(c)
{
}

MapDbView::~MapDbView() {
}

ClassId MapDbView::GetClassId() { return MAPDB_VIEW; }

boolean MapDbView::IsA(ClassId id) {
    return id == MAPDB_VIEW || OverlaysView::IsA(id);
}

MapLibView* MapDbView::FindLibView(MapLibrary* maplib) {
    Iterator i;
    for (First(i); !Done(i); Next(i)) {
	MapLibView* mlv = (MapLibView*)GetView(i);
	if (((MapLibComp*)mlv->GetGraphicComp())->GetLibrary() == maplib)
	    return mlv;
    }
    return nil;
}

void MapDbView::Interpret(Command* cmd) {
    if (cmd->IsA(MAPSHOW_CMD)) {
	MapLibrary* maplib = (MapLibrary*)((MapDbComp*)GetGraphicComp())->GetDatabase()->library(((MapShowCmd*)cmd)->GetLibName());
	if (maplib) {
	    MapLibView* mlv = FindLibView(maplib);
	    if (mlv)
		mlv->Interpret(cmd);
	}
    }
    else if (cmd->IsA(MAPHIDE_CMD)) {
	MapLibrary* maplib = (MapLibrary*)((MapDbComp*)GetGraphicComp())->GetDatabase()->library(((MapHideCmd*)cmd)->GetLibName());
	if (maplib) {
	    MapLibView* mlv = FindLibView(maplib);
	    if (mlv)
		mlv->Interpret(cmd);
	}
    }
    else OverlaysView::Interpret(cmd);
}

void MapDbView::Uninterpret(Command* cmd) {
}

/****************************************************************************/

MapEditor* MapIdrawView::_ed = nil;

MapIdrawView::MapIdrawView (MapIdrawComp* subj) : FrameIdrawView(subj) { 
}

ClassId MapIdrawView::GetClassId () { return MAP_IDRAW_VIEW; }

boolean MapIdrawView::IsA (ClassId id) {
    return MAP_IDRAW_VIEW == id || FrameIdrawView::IsA(id);
}
